#include <ros/ros.h>
#include <visualization_msgs/Marker.h>
#include <visualization_msgs/MarkerArray.h>
#include <laserNode/Obstacle.h>
#include <laserNode/Obstacles.h>
#include <std_srvs/Empty.h>
#include <GVG/Scan360.h>
#include <boost/lexical_cast.hpp>
#include <cmath>

ros::ServiceClient  scan_cln;
ros::Publisher obstacles_pub;
ros::Publisher obstacle_names_pub;
bool do_test(std_srvs::Empty::Request& req, std_srvs::Empty::Response& res); 
void handle_obstacles(const laserNode::Obstacles& msg);

int main( int argc, char** argv ) {
  ros::init(argc, argv, "scan360_viz");
  ros::NodeHandle nh;
  obstacle_names_pub = nh.advertise<visualization_msgs::MarkerArray>("obstacle_names", 1);
  obstacles_pub = nh.advertise<visualization_msgs::MarkerArray>("obstacles", 1);
  scan_cln      = nh.serviceClient<GVG::Scan360>("/indoor/gvg/scan360"); 
  ros::ServiceServer  do_scan  = nh.advertiseService("/indoor/gvg/scan360/do_test", &do_test); 
  ros::spin();
  return 0;
}

bool do_test(std_srvs::Empty::Request& req, std_srvs::Empty::Response& res) {
  GVG::Scan360 srv;
  srv.request.ang_vel = 20;
  scan_cln.call(srv);
  handle_obstacles(srv.response.obstacles);
  return true;
}
 
void handle_obstacles(const laserNode::Obstacles& msg) {
  std::string ns = "closest_obstacles";
  std::string frame_id = "/base_laser_link";
  visualization_msgs::MarkerArray ma;
  
  /*
   * Populating obstacles
   */
  for (int i = 0; i < (int) msg.collection.size(); i++) {
    visualization_msgs::Marker m;
    m.header.frame_id = frame_id;
    m.header.stamp = ros::Time::now();
    m.ns = ns;
    m.id = i;
    m.type = visualization_msgs::Marker::LINE_STRIP;
    m.action = visualization_msgs::Marker::ADD;
    m.scale.x = 0.03;  
    
    if (i % 2 == 0) {
      m.color.a = 1.0;   m.color.r = 0.0;   m.color.g = 1.0;  m.color.b = 0.0;
    } else {
      m.color.a = 1.0;   m.color.r = 0.0;   m.color.g = 0.0;  m.color.b = 1.0;
    }

    std::vector<geometry_msgs::Point> pts;
    for (int j = 0; j < (int) msg.collection.at(i).surface.size(); j++) {
      geometry_msgs::Point pt;
      pt.x = msg.collection.at(i).surface.at(j).x;
      pt.y = msg.collection.at(i).surface.at(j).y;
      pts.push_back(pt);
    }
    
    m.points.assign(pts.begin(), pts.end());
    m.lifetime = ros::Duration(20.0);   //20 seconds
    ma.markers.push_back(m);
  }
  
  /*
   * Populating names of obstacles
   */
  visualization_msgs::MarkerArray ma_names;
  for (int i = 0; i < (int) msg.collection.size(); i++) {
    visualization_msgs::Marker m;
    m.header.frame_id = frame_id;
    m.header.stamp = ma.markers.at(i).header.stamp;
    m.ns = "obstacle_names";
    m.id = i;
    m.type = visualization_msgs::Marker::TEXT_VIEW_FACING;
    m.action = visualization_msgs::Marker::ADD;
    m.scale.z = 0.4;  
    m.text = boost::lexical_cast<std::string>(i);
    m.lifetime = ma.markers.at(i).lifetime;
    m.color = ma.markers.at(i).color;
    int size = (int) msg.collection.at(i).surface.size();
    m.pose.position.x = msg.collection.at(i).surface.at(floor(size/2.0)).x;
    m.pose.position.y = msg.collection.at(i).surface.at(floor(size/2.0)).y;
    m.pose.position.z = msg.collection.at(i).surface.at(floor(size/2.0)).z;
    ma_names.markers.push_back(m);
  }

  obstacles_pub.publish(ma);
  obstacle_names_pub.publish(ma_names);
}

