#ifndef LASER_UTILS_H
#define LASER_UTILS_H

#include <vector>
#include <string>
#include <ros/ros.h>
#include <laserNode/Obstacle.h>
#include <laserNode/Obstacles.h>
#include <geometry_msgs/Point32.h>
#include <sensor_msgs/LaserScan.h>

/* Minimum distance that is considered a scan discontinuity between laser scans. */
#define MIN_RANGE_JUMP 0.5

/* Defines the max size of a scan window around a certain scan */
#define WINDOW_SIZE 30     

/* Defines the max number of adjacent laser scans that will be considered 
   when seeing if a particular scan point corresponds to a corner */
#define CORNER_WINDOW 30

/* Lines that have this orientation difference will be considered to 
   have no orientation difference at all. */
#define SAME_LINE_ANGLE 30

/* Minimum bearing difference between the two obstacles that will be used in the
   GVG midline definition, while following a GVG edge. */
#define SAME_OBJECT_MIN_BEARING 50

/* Max distance difference between two consecutive laser scans that belong to the same object [in m]. */
#define MAX_DIST_SAME_OBS_SCANS 0.1

#define ROBOT_DIAM 0.5  // 40cm

class LaserUtils {

 public:
  LaserUtils(std::string& robot_name);
  
  void   laserScanFiltering(const sensor_msgs::LaserScan::ConstPtr& msg);

  void   handleLaserScan(const sensor_msgs::LaserScan::ConstPtr& msg); 
    
  void   getClosestObstaclesViaAnnulus(laserNode::Obstacles& annulus_obstacles, double epsilon, 
				       const sensor_msgs::LaserScan::ConstPtr& msg);
 private:
  double getRange(const sensor_msgs::LaserScan::ConstPtr& msg, int i); 
  double getBearing(const sensor_msgs::LaserScan::ConstPtr& msg, int i); 
  geometry_msgs::Point32 toPoint32(const sensor_msgs::LaserScan::ConstPtr& msg, int i);
  
  ros::NodeHandle     nh;
  ros::Subscriber     laserSub;
  ros::Subscriber     filteredLaserSub;
  ros::Publisher      filteredScanPub;
  ros::Publisher      allObsPub;
  int                 countScans;
 
  
};

bool   areSameObject(laserNode::Obstacle& ob1, laserNode::Obstacle& ob2);
void   mergeObstacles(laserNode::Obstacle& ob1, laserNode::Obstacle& ob2);
#endif
