#include <ros/ros.h>
#include <robotNode/Follow_Wall.h>
#include <laserNode/Obstacles.h>
#include <laserNode/Obstacle.h>
#include <string>
#include <cmath>

ros::Publisher wf_pub;

void    handle_obstacles(const laserNode::Obstacles::ConstPtr& msg);

double  error_in_theta(geometry_msgs::Point32& closest_point) {
  double bearing_from_x_axis = 180.0/M_PI * atan2(closest_point.y, closest_point.x);
  return bearing_from_x_axis + 90; // following right wall
}

int main(int argc, char **argv) {
  ros::init(argc, argv, "test_wall_following");
  ros::NodeHandle nh;
  
  wf_pub = nh.advertise<robotNode::Follow_Wall>("/indoor/follow_wall", 1);
  ros::Subscriber obstacles_sub = nh.subscribe("/indoor/laser_utils/closest_obstacles", 1, &handle_obstacles);
  ros::spin();
  return 0;
}

void handle_obstacles(const laserNode::Obstacles::ConstPtr& msg) {
  // compute dy, dtheta and send the wall-follow message
  if (msg->collection.empty()) {
    return;
  }

  laserNode::Obstacle     closest = msg->collection.at(0);
  geometry_msgs::Point32  closest_point = closest.closest_point;
  robotNode::Follow_Wall m;
  m.dtheta_in_deg = error_in_theta(closest_point);
  m.dy_in_m = 1 - closest.min_distance;
  m.lin_vel = 1;
  m.range_of_closest_point = closest.min_distance;
  m.theta_of_closest_point = 180.0/M_PI * atan2(closest_point.y, closest_point.x);
  
  ROS_INFO("dtheta %f", m.dtheta_in_deg);
  ROS_INFO("dy %f", m.dy_in_m);
  ROS_INFO("theta %f", m.theta_of_closest_point);
  ROS_INFO("y %f", m.range_of_closest_point);
  
  wf_pub.publish(m);
}
