#include <iostream>
#include <limits>
#include "localizerGVG.h"
#include <tf/transform_listener.h>


localizerGVG::localizerGVG(std::string& robot_name):
  X(3),P(3,3) {
   ROS_INFO("localizerGVG::localizerGVG ");

  std::string output_frame;
  std::string theMap_frame;

  odomSub = nh.subscribe("/indoor/odom",1,&localizerGVG::handleOdom, this);
  startService = nh.advertiseService("StartFilter"    , &localizerGVG::processStart,this);
  updateService = nh.advertiseService("UpdateFilter"  , &localizerGVG::processMeetpoint,this);
  maxUncService = nh.advertiseService("MaxUncertainty", &localizerGVG::processMaxUnc,this);
  minUncService = nh.advertiseService("MinUncertainty", &localizerGVG::processMinUnc,this);

  nh.param("output_frame", output_frame, std::string("odom_combined"));
  posePub = nh.advertise<nav_msgs::Odometry>(output_frame, 10);

   nh.param("theMap_frame", theMap_frame, std::string("theMap"));

  mapPub = nh.advertise<localizer::GVGmap>(theMap_frame, 10);
  
  filterOn=true;
  initOdom=false;
  nL=0; // number of landmarks
  // EKF constants
  X=VectorXd(3); // The state vector initialized to the pose of the robot.
  X<< 0,0,0;

  P=MatrixXd::Zero(3,3); // The covariance matrix

  // The model noise covariance (linear and angular velocity)
  Qr<< _Wvv, _Wvw,
        _Wvw, _Www;


  CF<<1,0,0,
    0,1,0,
    0,0,1;

  J<<0, -1,
    1, 0;

  //cout<<"J"<<J<<endl;

  R<< 0.02527455743763,   0.000014869773007,0,
    0.000014869773007,   0.02476827355463,0,
    0,0,0.025;

}


void localizerGVG::handleOdom(const nav_msgs::Odometry::ConstPtr& odom) {
//void localizerGVG::handleOdom(const geometry_msgs::PoseWithCovarianceStamped::ConstPtr& odom) {

  double x=odom->pose.pose.position.x;
  double y=odom->pose.pose.position.y;
  //  double yaw=acos(odom->pose.pose.orientation.w)*2.0;
  tf::Pose pose;
  tf::poseMsgToTF(odom->pose.pose, pose);
  double yaw=tf::getYaw(pose.getRotation());
   // ROS_INFO("localizerGVG::handleOdom received [%lf,%lf,%lf]",x,y,yaw*180/M_PI);
  if(!initOdom) {
    oldX=x;
    oldY=y;
    oldYaw=yaw;
    initOdom=true;
    oldStamp=odom->header.stamp;
    X<<x,y,yaw;
  //  ROS_INFO("localizerGVG::handleOdom X initialized to [%lf,%lf,%lf]",X(0),X(1),X(2)*180/M_PI);
    return;
  }
  if(filterOn) {
    if((oldX==x)&&(oldY==y)&&(oldYaw==yaw)){
      oldStamp=odom->header.stamp;
      //      ROS_INFO("localizerGVG::handleOdom same position ending");
      return;
    }
    // Calculate the measurement
    double dx=x-oldX;
    double dy=y-oldY;
    double dYaw=angleDiff(yaw,oldYaw);
    double dt=odom->header.stamp.toSec()-oldStamp.toSec();
    //ROS_INFO("localizerGVG::handleOdom dx,dy,dYaw,dt=[%lf,%lf,%lf,%lf]",dx,dy,dYaw,dt);
    double Vm=sqrt(dx*dx+dy*dy)/dt;
    double Wm=dYaw/dt;
    //    ROS_INFO("localizerGVG::handleOdom 1");
    // Perform Propagate
    //ROS_INFO("localizerGVG::handleOdom Vm dt X(2) cos(X(2)) sin(X(2)) %lf %lf %lf %lf %lf", Vm, dt, X(2), cos(X(2)), sin(X(2))); 
    //ROS_INFO("localizerGVG::handleOdom X0 dx v*dt*cos(a) %lf %lf %lf",X(0),dx,Vm*dt*cos(X(2)));
    //ROS_INFO("localizerGVG::handleOdom X1 dy v*dt*sin(a) %lf %lf %lf",X(1),dy,Vm*dt*sin(X(2)));
    // ROS_INFO("localizerGVG::handleOdom x X(0) x+dx oldX+Vm*dt*cos(X(2)) %lf %lf %lf %lf", x, X(0),x+dx, oldX+Vm*dt*cos(X(2)));
    // ROS_INFO("localizerGVG::handleOdom y X(1) y+dy oldY+Vm*dt*sin(X(2)) %lf %lf %lf %lf", y, X(1),y+dy, oldY+Vm*dt*sin(X(2)));
    X(0)=X(0)+Vm*dt*cos(X(2));
    X(1)=X(1)+Vm*dt*sin(X(2));
    X(2)=thetapp(X(2)+Wm*dt);
 //       ROS_INFO("localizerGVG::handleOdom X  [%lf,%lf,%lf]",X(0),X(1),X(2)*180/M_PI);
    //   ROS_INFO("localizerGVG::handleOdom 2");
    propagate(Vm,Wm,dt);
    // Publish the odometry with covariance topic
    geometry_msgs::Quaternion odom_quat = tf::createQuaternionMsgFromYaw(X(2));
    nav_msgs::Odometry outputPose;
    outputPose.header.frame_id = "odom";
    outputPose.header.stamp =odom->header.stamp;
    outputPose.pose.pose.position.x = X(0);
    outputPose.pose.pose.position.y = X(1);
    outputPose.pose.pose.position.z = 0;
    outputPose.pose.pose.orientation=odom_quat;
    for (unsigned int i=0; i<2; i++)
      for (unsigned int j=0; j<2; j++)
    	outputPose.pose.covariance[6*i+j] = P(i,j);
    posePub.publish(outputPose);
  }
  oldX=x;
  oldY=y;
  oldYaw=yaw;
  //cout<<"OLD X-----"<<oldX<<" "<<oldY<<" "<<oldYaw<<endl;
  oldStamp=odom->header.stamp;
  //  ROS_INFO("localizerGVG::handleOdom ending");
}

void localizerGVG::propagate(double Vm, double Wm, double dt) {
  //ROS_INFO("localizerGVG::propagate start ");
  Matrix3d Fr;
  Fr<<1, 0, -Vm*dt*sin(oldYaw),
      0, 1,  Vm*dt*cos(oldYaw),
      0, 0,  1;
  MatrixXd Gr(3,2);
  Gr<< -dt*cos(oldYaw), 0, 
       -dt*sin(oldYaw), 0,
        0,             -dt;
    
  //Pr=Fr*Pr*Fr.transpose()+Gr*Qr*Gr.transpose();
  P.block(0,0,3,3)=Fr*P.block(0,0,3,3)*Fr.transpose()+Gr*Qr*Gr.transpose();
  CF=Fr*CF;
  normalizeCovariance();
  //ROS_INFO("localizerGVG::propagate end ");
}



bool localizerGVG::processStart(localizer::StartFilter::Request  &req,
			localizer::StartFilter::Response &res) {
  if(req.start) {
    ROS_INFO("localizerGVG::StartFilter received true");
  }
  else  {
    ROS_INFO("localizerGVG::StartFilter received false");
  }
  filterOn=req.start;
  return true;
}


void localizerGVG::propagateRL(){
// Propagate the cross-corelation between robot and landmarks up to now
  // ROS_INFO("localizerGVG::propagateRL start");
//  ROS_INFO("localizerGVG::propagateRL nL %d",nL);
  // cout<<P.size()<<endl;
  for(int i=0; i<nL; i++) {
    P.block(0,3+i*3,3,3)=CF*P.block(0,3+i*3,3,3);
    P.block(3+i*3,0,3,3)=P.block(3+i*3,0,3,3)*CF.transpose();
  }
  CF<<1,0,0,
    0,1,0,
    0,0,1;
  // ROS_INFO("localizerGVG::propagateRL end");
}

bool localizerGVG::processMeetpoint(localizer::UpdateFilter::Request  &req,
				    localizer::UpdateFilter::Response &res) {

  ROS_INFO("localizerGVG::processMeetpoint received meetpoint %d [%lf,%lf,%lf]", req.id, req.x, req.y,req.yaw);
  ROS_INFO("localizerGVG::processMeetpoint robot pose [%lf,%lf,%lf]",X(0),X(1),X(2));
  if(nL==0) {
    //Reset the odometry covariance.
    P=MatrixXd::Zero(3,3);
  }
  // Propagate the cross-corelation between robot and the 
  // existing landmarks up to now
  propagateRL();

  // For the new landmark
  ROS_INFO("localizerGVG::processMeetpoint Prelimin");

  Vector3d xl; //the coordinates of the landmark (meetpoint) in World Frame
  Vector3d z; // Vector representation of the measurment
  z<<req.x, req.y, req.yaw;
  xl.segment(0,2)=X.segment(0,2)+C(X(2))*z.segment(0,2);
  xl(2)=X(2)+z(2);
  // Create the measurment matrices
  MatrixXd Hr(3,3);
  Vector2d dx;
  dx=xl.segment(0,2)-X.segment(0,2);
  double _c=cos(X(2));
  double _s=sin(X(2));
  Hr<<-_c, -_s, + _c*dx(1)-_s*dx(0),
    _s, -_c, - _s*dx(1)-_c*dx(0),
    0,   0, -1;
  Matrix3d Hl; // Hl is a 2 by 2 Matrix
  Hl=MatrixXd::Zero(3,3);
  Hl.block(0,0,2,2)=CT(X(2));
  Hl(2,2)=1;
  ROS_INFO("localizerGVG::processMeetpoint Check for id %d",req.id);
  if(newLandmark(req.id)) {
    // New Landmark
    //  ROS_INFO("localizerGVG::processMeetpoint New Landmark");
    idList.push_back(req.id);
    ROS_INFO("localizerGVG::processMeetpoint  Expand");
    P.conservativeResize(P.rows()+3,P.cols()+3);
    X.conservativeResize(X.size()+3);
    X.segment(3+nL*3,3)=xl;
    ROS_INFO("localizerGVG::processMeetpoint Create new Pllcv");
    // Create new Pllcv matrices describing the cross corelation 
    // between the old landmarks and the new Landmark. Page 79 eq. 5.73 
    for (int count=0; count<nL; count++){
      //  ROS_INFO("localizerGVG::processMeetpoint Pll for landmark %d %d",count, req.id);
      Matrix3d newPllcv;
      newPllcv=-P.block(3+count*3,0,3,3)*Hr.transpose()*Hl;
      P.block(3+count*3,3+nL*3,3,3)=newPllcv;
      P.block(3+nL*3,3+count*3,3,3)=newPllcv.transpose();    
    }
    // Create new Prl matrix describing the cross corelation 
    // between the Robot and the new Landmark. Page 79 eq. 5.73 
    ROS_INFO("localizerGVG::processMeetpoint Create new Prl matrix");
    MatrixXd newPrl(3,3);
  
    // cout<<-P.block(0,0,3,3)<<endl;
    // cout<<Hr.transpose()<<endl;
    // cout<<Hl<<endl;
    
    newPrl=-P.block(0,0,3,3)*Hr.transpose()*Hl;
    P.block(0,3+nL*3,3,3)=newPrl;
    P.block(3+nL*3,0,3,3)=newPrl.transpose();
    // Create the new Pll matrix describing the covariance of the landmark
    // page 79 eq. 5.74
    ROS_INFO("localizerGVG::processMeetpoint Create the new Pll matrix");
    Matrix3d newPll;
    newPll=Hl.transpose()*(Hr*P.block(0,0,3,3)*Hr.transpose()+R)*Hl;
    P.block(3+nL*3,3+nL*3,3,3)=newPll;
    nL++;
  }
  else {
    // Find which landmark it is:
    int id=-1;
    for(unsigned int i=0; i<idList.size(); i++){
      if(idList[i]==req.id)
	id=i;
    }
    ROS_INFO("localizerGVG::processMeetpoint id %d %d",req.id, id);
    //Sanity check:
    if(id<0) {
      cerr<<"Not able to find id:"<<req.id<< "in the list of ids"<<endl;
      for(unsigned int i=0; i<idList.size(); i++){
	cerr<<idList[i]<<" ";
      }
      cerr<<endl;
    }
    // perform update

    Vector3d z_est;
    z_est.segment(0,2)=CT(X(2))*(X.segment(3+3*id,2)-X.segment(0,2));
    z_est(2)=X(3+3*id+2)-X(2);

    Vector3d r;
    r.segment(0,2)=z.segment(0,2)-z_est.segment(0,2);
    r(2)=angleDiff(z(2),z_est(2));
    cout<<"z     "<<z<<endl;
    cout<<"Z_est "<<z_est<<endl;
    cout<<"r     "<<r<<endl;
    MatrixXd H(3,3+3*nL);
    H=MatrixXd::Zero(3,3+3*nL);

    H.block(0,0,3,3)=Hr;
    H.block(0,3+3*id,3,3)=Hl;
    MatrixXd S(3,3);
    S=H*P*H.transpose()+R;
    MatrixXd K(3,3+3*nL);
    K=P*H.transpose()*S.inverse();
    X=X+K*r;
    X(2)=thetapp(X(2));
    P=(MatrixXd::Identity(3+nL*3,3+nL*3)-K*H)*P*(MatrixXd::Identity(3+nL*3,3+nL*3)-K*H).transpose()+K*R*K.transpose();
  }
  // publish the odometry with covariance topic    
  
  // Publish the Map
  //publishMap();
  localizer::GVGmap theMap;
  //theMap.frame_id = "map";
  //theMap.stamp =oldStamp;
  cout<<"+++++++++ "<<X.size()<<" ";
  for(unsigned int i=0; i<X.size(); i++) {
    theMap.state.push_back(X[i]);
    res.X.push_back(X[i]);
    cout<<X[i]<<" ";
  }
  cout<<endl;

  for(unsigned int i=0; i<P.rows(); i++) {
    for(unsigned int j=0; j<P.cols(); j++) {
      theMap.cov.push_back(P(i,j));
      res.P.push_back(P(i,j));
    }
  }
  mapPub.publish(theMap);
  // cerr<<"The id list:\n";
  // cerr<<"--------------------------------------------------------------\n";
  // for(unsigned int i=0; i<idList.size(); i++){
  //   cerr<<idList[i]<<" ";
  // }
  // cerr<<endl;
  ROS_INFO("localizerGVG::processMeetpoint end");
  return(true);
}
double localizerGVG::thetapp(double theta) {
//cout<<"-------------------------------------------- "<<theta;
  while(theta>M_PI) theta-=2.0*M_PI;
  while(theta<-M_PI) theta+=2.0*M_PI;
return(theta);
}
double localizerGVG::theta02p(double theta) {
  while(theta>2*M_PI) theta-=2.0*M_PI;
  while(theta<0) theta+=2.0*M_PI;
return(theta);
}

bool localizerGVG::newLandmark(int id){
  // ROS_INFO("localizerGVG::newLandmark");
 for (vector<int>::iterator it = idList.begin(); it != idList.end(); ++it){
    if((*it)==id) return(false);
  }
  // ROS_INFO("localizerGVG::newLandmark");
  return true;
}

// insideEllipse() {
// for ellipse with a and b semi axis oriented by angle t:
// ((x-xc)*cos(t)-(y-yc)*sin(t)).^2/a^2 + ...
// ((x-xc)*sin(t)+(y-yc)*cos(t)).^2/b^2 <= 1
//}

double localizerGVG::angleDiff(double a, double b) {
  double diff2= theta02p(a)-theta02p(b);
  while (diff2>M_PI)  diff2-=2*M_PI; 
  while (diff2<-M_PI) diff2+=2*M_PI;
  return(diff2);
}



void localizerGVG::normalizeCovariance(){
  for(int i=0; i<P.rows()-1; i++) {
    for (int j=i+1; j<P.cols(); j++){      
      double tmp=(P(i,j)+P(j,i))/2.0;
      P(i,j)=tmp;
      P(j,i)=tmp;
    }
  }
}

Matrix2d localizerGVG::CT(double angle) {
  Matrix2d c;
  c<<cos(angle), sin(angle),
    -sin(angle), cos(angle);
  return(c);
}

Matrix2d localizerGVG::C(double angle) {
  Matrix2d c;
  c<<cos(angle), -sin(angle),
    sin(angle), cos(angle);
  return(c);
}

bool localizerGVG::processMaxUnc(localizer::MaxUncertainty::Request  &req,
				 localizer::MaxUncertainty::Response &res)  {
  double maxT=0;
  res.id=-1;
  for(int i=0; i<nL; i++) {
    double tmpT=P.block(3+i*2,3+i*2,2,2).trace();
    if(tmpT>maxT) {
      maxT=tmpT;
      res.id=i;
    }
  }
  return true;
  
}

bool localizerGVG::processMinUnc(localizer::MinUncertainty::Request  &req,
				 localizer::MinUncertainty::Response &res)  {
  double minT=1000000;
  res.id=-1;
  for(int i=0; i<nL; i++) {
    double tmpT=P.block(3+i*2,3+i*2,2,2).trace();
    if(tmpT<minT) {
      minT=tmpT;
      res.id=i;
    }
  }
  return true;
  
}


//////////////////////////////////////////////////////////////
// Legacy code, first attempt using submatrices 
// and also using two different calls for augment and update.
//////////////////////////////////////////////////////////////

// bool localizerGVG::processAugment(localizer::AugmentFilter::Request  &req,
// 			localizer::AugmentFilter::Response &res) {
//   ROS_INFO("localizer::processAugment received meetpoint %d [%lf,%lf]", req.id, req.x, req.y);

//   // Propagate the cross-corelation between robot and landmarks up to now
//   propagateRL();

//   // For the new landmark
  
//   Vector2d xl; //the coordinates of the landmark (meetpoint) in World Frame
//   //Vector2d xr;
//   Vector2d z; // Vector representation of the measurment
//   //xr<<oldX, oldY;
//   z<<req.x, req.y;
//   xl=X.segment(0,2)+C(X(2))*z;

//   // Create the measurment matrices
//   MatrixXd Hr(2,3);
//   Vector2d dx;
//   dx=xl-X.segment(0,2);
//   double _c=cos(X(2));
//   double _s=sin(X(2));

//   Hr<<-_c, -_s, + _c*dx(1)-_s*dx(0),
//        _s, -_c, - _s*dx(1)-_c*dx(0);

//   Matrix2d Hl; // Hl is a 2 by 2 Matrix
//   Hl=CT(X(2));

//   P.conservativeResize(P.rows()+2,P.cols()+2);
//   X.conservativeResize(X.size()+2);
//   X.segment(3+nL*2,2)=xl;
//   // Create new Pllcv matrices describing the cross corelation 
//   // between the old landmarks and the new Landmark. Page 79 eq. 5.73 
//   // for (vector<MatrixXd>::iterator it = Prl.begin(), int count=0; it != Prl.end(); ++it, count++){
//   //   Matrix2d newPllcv;
//   //   newPllcv=-(*it)*Hr.transpose()*Hl;
//   //   P.block(3+count*2,3+nL*2,2,2)=newPllcv;
//   //   P.block(3+nL*2,3+count*2,2,2)=newPllcv.transpose();
//   //   //    Pllcv.push_back(newPllcv);
    
//   // }
//   for (int count=0; count<nL; count++){
//     Matrix2d newPllcv;
//     newPllcv=-P.block(3+count*2,0,3,2)*Hr.transpose()*Hl;
//     P.block(3+count*2,3+nL*2,2,2)=newPllcv;
//     P.block(3+nL*2,3+count*2,2,2)=newPllcv.transpose();    
//   }
//   // Create new Prl matrix describing the cross corelation 
//   // between the Robot and the new Landmark. Page 79 eq. 5.73 

//   MatrixXd newPrl(3,2);
//   newPrl=-Pr*Hr.transpose()*Hl;
//   P.block(0,3+nL*2,3,2)=newPrl;
//   P.block(3+nL*2,0,3,2)=newPrl.transpose();
//   // Prl.push_back(newPrl);

//   // Create the new Pll matrix describing the covariance of the landmark
//   // page 79 eq. 5.74
//   Matrix2d newPll;
//   newPll=Hl.transpose()*(Hr*Pr*Hr.transpose()+R)*Hl;
//   P.block(3+nL*2,3+nL*2,2,2)=newPll;
//   //Pll.push_back(newPll);
//   nL++;
//   return true;
// }
// bool localizerGVG::processUpdate1(localizer::UpdateFilter::Request  &req,
// 			localizer::UpdateFilter::Response &res) {
//   ROS_INFO("localizer::processUpdate received meetpoint %d [%lf,%lf]", req.id, req.x, req.y);
// // Propagate the cross-corelation between robot and landmarks up to now
//   propagateRL();
//   // Calculate measurement

//   // Reconstructing the covariance matrix:

//   int size=3+2*nL;
//   MatrixXd _P(size,size);
//   _P.block<3,3>(0,0)=Pr;
//   int count=0;
//   for (vector<Matrix2d>::iterator it = Pll.begin(); it != Pll.end(); ++it) {
//     _P.block(3+count*2,3+count*2,2,2)=(*it);
//   }

//   vector<Matrix2d>::iterator it = Pllcv.begin();
//   for(int i=0; i<nL; i++) {
//     for(int j=0; j<i; j++) {
//       _P.block(3+(i+1)*2,3+j*2,2,2)=(*it);
//       _P.block(3+j*2,3+(i+1)*2,2,2)=(*it).transpose();
//       it++;
//     }
//   }

//   vector<MatrixXd>::iterator itx = Prl.begin();
//   for(int i=0; i<nL; i++) {
//     _P.block(3+i*2,0,2,3)=(*itx);
//     _P.block(0,3+i*2,3,2)=(*itx).transpose();
//     itx++;
//   }

//   // perform update

//   // publish the odometry with covariance topic

//   // Reconstruct the sub-matrices after the update
//   Pr=_P.block<3,3>(0,0);
//   count=0;
//   for (vector<Matrix2d>::iterator it = Pll.begin(); it != Pll.end(); ++it) {
//     (*it)=_P.block(3+count*2,3+count*2,2,2);
//   }

//   itx = Prl.begin();
//   for(int i=0; i<nL; i++) {
//     (*itx)=_P.block(3+i*2,0,2,3);
//     itx++;
//   }

//  return true;

// }
