#include <vector>
#include <string>
#include <Eigen/Dense>
#include <cstdlib>
#include <iostream>
using namespace Eigen;
using namespace std;
    
Matrix2d CT(double angle) {
  Matrix2d CT;
  CT<<cos(angle), sin(angle),
    -sin(angle), cos(angle);
  return(CT);

}

int main(int argc, char **argv) {

  cout<<"Testing Eigen"<<endl;
  Vector2d dx;
  VectorXd X(5);
  X=VectorXd::Random(5);
  X(2)=1.1441;
  cout<<"X \n"<<X<<endl;
  cout<<"--------"<<endl;


  Vector2d xl;
  xl=Vector2d::Random(2);
  cout<<"xl \n"<<xl<<endl;
  cout<<"--------"<<endl;

  dx=xl-X.segment(0,2);
  double _c=cos(X(2));
  double _s=sin(X(2));

  MatrixXd Hr(2,3);
  Hr<<-_c, -_s, + _c*dx(1)-_s*dx(0),
    _s, -_c, - _s*dx(1)-_c*dx(0);

  cout<<"Hr \n"<<Hr<<endl;
  cout<<"--------"<<endl;
  Matrix2d Hl; 
  Hl=CT(1.1441);
  cout<<"Hl \n"<<Hl<<endl;
  cout<<"--------"<<endl;
  cout<<"Hr.transpose*Hl \n"<<Hr.transpose()*Hl<<endl;
  cout<<"--------"<<endl;

  cout<<"------------------------------------------"<<endl;

  MatrixXd P(5,5);
  P=MatrixXd::Random(5,5);


  cout<<"Testing Prl"<<endl;
  cout<<"Prr \n"<<P.block(0,0,3,3)<<endl;
  cout<<"--------"<<endl;
  cout<<"Prr*Hr^T*Hl \n"<<P.block(0,0,3,3)*Hr.transpose()*Hl<<endl;
  cout<<"------------------------------------------"<<endl;
  cout<<"Testing Propagate rl"<<endl;  
  Matrix3d CF;
  MatrixXd Prl(3,2);
  CF=MatrixXd::Random(3,3);
  Prl=MatrixXd::Random(3,2);
  cout<<CF*Prl<<endl;
  cout<<Prl.transpose()*CF.transpose()<<endl;
  cout<<"------------------------------------------"<<endl;

  int nL=1;
  for(int i=0; i<nL; i++) {
    cout<<CF*P.block(0,3+i*2,3,2)<<endl;
    cout<<P.block(3+i*2,0,2,3)*CF.transpose()<<endl;
  }

  cout<<"------------------------------------------"<<endl;
  cout<<"Testing create P ll"<<endl;

  P.conservativeResize(P.rows()+2,P.cols()+2);

  cout<<"P dimensions "<<P.rows()<<" by "<<P.cols()<<endl;

  for (int count=0; count<nL; count++){
    cout<<"Prl\n"<<P.block(3+count*2,0,2,3)<<endl;
    cout<<"-------------------------"<<endl;

    cout<<"Hr^T\n"<<Hr.transpose()<<endl;
    cout<<"-------------------------"<<endl;

    cout<<"Hl\n"<<Hl<<endl;

    cout<<"-------------------------"<<endl;
    //cout<<-P.block(3+count*2,0,2,3)*Hr.transpose()<<endl;
    cout<<"Hr^T*Hl\n"<<(Hr.transpose()*Hl)<<endl;
    cout<<"-------------------------"<<endl;

    cout<<"Prl*Hr^T*Hl\n"<<-P.block(3+count*2,0,2,3)*(Hr.transpose()*Hl)<<endl;
    cout<<"---------------------------"<<endl;
    Matrix2d newPllcv;
    newPllcv=-P.block(3+count*2,0,2,3)*Hr.transpose()*Hl;
    P.block(3+count*2,3+nL*2,2,2)=newPllcv;
    P.block(3+nL*2,3+count*2,2,2)=newPllcv.transpose();    
  }
 



  VectorXd v(3);
  v<< 1,2,3;

  MatrixXd M(6,3);
  M<< 1, 2, 3, 1, 2, 3,
    4, 5, 6, 4, 5, 6,
    7, 8, 9, 7, 8, 9;
  cout<<"M="<<M<<endl;

  M.conservativeResize(M.rows()+2,M.cols()+2);
  cout<<"M="<<M<<endl;

  //  MatrixXd P(11,11);
  //  P=MatrixXd::Zero(11,11);

  //  P.block(0,0,3,3)=MatrixXd::Constant(3,3,9);
  //  cout<<"P=\n"<<P<<endl;
  //  for(int i=0; i<4; i++) {
  //    P.block(3+i*2,3+i*2,2,2)=MatrixXd::Constant(2,2,i+1); // LiLi component
  //    P.block(0,3+i*2,3,2)=MatrixXd::Constant(3,2,9+i+1); //RL component
  //    // To Do LR component 


  //  }
  //  cout<<"P=\n"<<P<<endl;

  // for(int i=0; i<11; i++) cout<<P(0,i)<<" ";
  // cout<<endl;


 
}
