#ifndef ROBOTSIM_H
#define ROBOTSIM_H

#include <ros/ros.h>
#include <robotNode/RelRotate.h>
#include <robotNode/Brake.h>
#include <robotNode/RelTranslate.h>
#include <robotNode/RelGoto.h>
#include <robotNode/Follow_Wall.h>
#include <nav_msgs/Odometry.h>
#include <string>
#include "PIDController.h"

class robotSim {

 public:
  
  robotSim(std::string& name, int id); 
  ~robotSim(); 

  /* Return the id of the robot, in case there are many */
  int getID() { return id; }

  /* Accepts dx, dy, dtheta as errors and drives the robot to a direction
     that will minimize those errors */
  void pdmove(const robotNode::RelGoto::ConstPtr& msg);
  
  /* Accepts dy, dtheta as errors and drives the robot to a direction
     that will minimize those errors */
  void follow_wall(const robotNode::Follow_Wall::ConstPtr& msg);
  

  /* Reads up the odometry of the robot from Stage  */
  void handleOdometry(const nav_msgs::Odometry::ConstPtr& msg);
  
  /* Motion command: tVel, the translational velocity is in m/sec 
     and aVel, the angular velocity is in rad/sec */ 
  void move(double tVel, double aRadVel);
  
  /* Move forward by dist meters */
  bool relTranslate(robotNode::RelTranslate::Request& req, 
		    robotNode::RelTranslate::Response& res);

  /* Rotate by dtheta in radians in place */
  bool relRotate(robotNode::RelRotate::Request& req, 
		 robotNode::RelRotate::Response& res);
			
  /* Stop the robot */
  bool stop(robotNode::Brake::Request& req,
	    robotNode::Brake::Response& res);
    
  double distance(geometry_msgs::Point& a, geometry_msgs::Point& b);

 private:
  int id;
  std::string name;
  nav_msgs::Odometry odom;
  PIDController *ang_vel_control;
  PIDController *dy_control;

  ros::NodeHandle    nh;
  ros::Publisher     movePub;
  ros::Subscriber    gotoSub;
  ros::Subscriber    odomSub;
  ros::Subscriber    wfSub;
  ros::ServiceServer relRotSrv;
  ros::ServiceServer relTranSrv;
  ros::ServiceServer brakeSrv;
};

#endif	// ROBOTSIM_H
